library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

use work.pkg_bwidow.all;
use work.pkg_d2a.all;

entity bwidow_xlink is
    Port ( O_PWM_XH : out  STD_LOGIC;
           O_PWM_XL : out  STD_LOGIC;
           O_PWM_YH : out  STD_LOGIC;
           O_PWM_YL : out  STD_LOGIC;
           O_VIDEO_Z : out  STD_LOGIC_VECTOR(3 downto 0);
           O_RDIS : out  STD_LOGIC;
           O_GDIS : out  STD_LOGIC;
           O_BDIS : out  STD_LOGIC;
			  O_DBG: out STD_LOGIC_VECTOR(15 downto 0);
			  I_PSXDATA				  : in    std_logic;
			  O_PSXCMD				  : out   std_logic;
			  O_PSXATT				  : out   std_logic;
           O_PSXCLK				  : out   std_logic;
	        I_PSXACK				  : in    std_logic;
           I_CLK : in  STD_LOGIC;
           I_RESET : in  STD_LOGIC;
           O_AUDIO_L : out  STD_LOGIC;
           O_AUDIO_R : out  STD_LOGIC);
end bwidow_xlink;


architecture Behavioral of bwidow_xlink is
signal buttons:		STD_LOGIC_VECTOR(14 downto 0);
signal psxbuttons:	STD_LOGIC_VECTOR(15 downto 0);
signal clk_ibuf:		STD_LOGIC;
signal clk_dlla_op0:	STD_LOGIC;
signal clk_dlla_div:	STD_LOGIC;
signal clk_dlla_2x:	STD_LOGIC;
signal reset_h:		STD_LOGIC;
signal reset_l:		STD_LOGIC;
signal clk_104:		STD_LOGIC;
signal clk_12:			STD_LOGIC;
signal clke_psx:		STD_LOGIC;
signal clkc_psx:		STD_LOGIC_VECTOR(3 downto 0);
signal clk_pwm:		STD_LOGIC;
signal xval:			STD_LOGIC_VECTOR(9 downto 0);
signal yval:			STD_LOGIC_VECTOR(9 downto 0);
signal xval_filt:			STD_LOGIC_VECTOR(9 downto 0);
signal yval_filt:			STD_LOGIC_VECTOR(9 downto 0);
signal xval_pwm:			STD_LOGIC_VECTOR(9 downto 0);
signal yval_pwm:			STD_LOGIC_VECTOR(9 downto 0);
signal xval_inv:			STD_LOGIC_VECTOR(9 downto 0);
signal yval_inv:			STD_LOGIC_VECTOR(9 downto 0);
signal zval	:			STD_LOGIC_VECTOR(7 downto 0);
signal rgb	:			STD_LOGIC_VECTOR(2 downto 0);
signal soundval:			STD_LOGIC_VECTOR(7 downto 0);
signal soundout:			STD_LOGIC;
signal cpu_reset_h:		STD_LOGIC;

attribute CLKDV_DIVIDE : string;
attribute CLKDV_DIVIDE of CLKDLLA : label is "4.5";
begin
	IBUFG0 : IBUFG port map (I=>I_CLK,  O=>clk_ibuf);
	CLKDLLA : CLKDLL port map (
		CLKIN  => clk_ibuf,
		RST    => reset_h,
		CLKFB  => clk_dlla_op0,
		CLK0   => clk_dlla_op0,
		CLKDV  => clk_dlla_div,
		CLK2X  => clk_dlla_2x
	);
	BUFG0 : BUFG port map (I=> clk_dlla_op0,O => clk_104);
	BUFG1 : BUFG port map (I=> clk_dlla_div,O => clk_12);
	BUFG2 : BUFG port map (I=> clk_dlla_2x,O => clk_pwm);

	pwmxh: pwmgen port map (
		inval => xval_pwm(9 downto 5),
		output => O_PWM_XH,
		clk => clk_pwm,
		reset => reset_h
	);
	pwmxl: pwmgen port map (
		inval => xval_pwm(4 downto 0),
		output => O_PWM_XL,
		clk => clk_pwm,
		reset => reset_h
	);
	pwmyh: pwmgen port map (
		inval => yval_pwm(9 downto 5),
		output => O_PWM_YH,
		clk => clk_pwm,
		reset => reset_h
	);
	pwmyl: pwmgen port map (
		inval => yval_pwm(4 downto 0),
		output => O_PWM_YL,
		clk => clk_pwm,
		reset => reset_h
	);
	
--	noisex: noisegen port map (
--		noise => xval,
--		clock => clk_12,
--		reset => reset_h
--	);
--	O_VIDEO_Z<="0010";

	filterx: filter generic map (
			mul => 1
		) port map (
			input => xval_inv,
			output => xval_filt,
			clk => clk_12,
			reset => cpu_reset_h
	);
	filtery: filter generic map (
			mul => 1
		) port map (
			input => yval_inv,
			output => yval_filt,
			clk => clk_12,
			reset => cpu_reset_h
	);
	audioda: deltasigma generic map (
			width => 8
		) port map (
			inval => soundval,
			output => soundout,
			clk => clk_pwm,
			reset => cpu_reset_h
	);

	mybwidow: bwidow port map (
		clk => clk_12,
		reset_h => cpu_reset_h,
		analog_sound_out => soundval,
		analog_x_out => yval,
		analog_y_out => xval,
		analog_z_out => zval,
		rgb_out => rgb,
		dbg => O_DBG,
		buttons => buttons
	);

	psxclk: process(clk_12, reset_h)
	begin
		if (reset_h='1') then
			clkc_psx<="0000";
		elsif rising_edge(clk_12) then
			clkc_psx<=clkc_psx+"0001";
			if (clkc_psx="0000") then
				clke_psx<='1';
			else
				clke_psx<='0';
			end if;
		end if;
	end process;

	psxcontroller : entity work.psxcontroller
		port map (
			psxdata => I_PSXDATA,
			psxcmd => O_PSXCMD,
			psxatt => O_PSXATT,
			psxclk => O_PSXCLK,
			psxack => I_PSXACK,
			buttons => psxbuttons,
			clk => clk_12,
			clken => clke_psx,
			reset_l => reset_l
		);

	--psxbuttons(15 downto 0): 
	--15:SQUARE 14:CROSS 13:CIRCLE 12:TRIANGLE 11:R1 10:L1 9:R2 8:L2 
	--7:LEFT 6:DOWN 5:RIGHT 4:UP 3:START 2:X 1:X 0:SELECT
	--buttons(14 downto 0): COINAUX COINL COINR START2 START1 FD FU FL FR MU MD ML MR
	buttons(14)<=not psxbuttons(10);
	buttons(13)<=not psxbuttons(8);
	buttons(12)<='1';
	buttons(11)<=not psxbuttons(11);
	buttons(10)<=not psxbuttons(9);
	buttons(9)<='1';
	buttons(8)<=not psxbuttons(3);
	buttons(7)<=not psxbuttons(12);
	buttons(6)<=not psxbuttons(14);
	buttons(5)<=not psxbuttons(15);
	buttons(4)<=not psxbuttons(13);
	buttons(3)<=not psxbuttons(4);
	buttons(2)<=not psxbuttons(6);
	buttons(1)<=not psxbuttons(7);
	buttons(0)<=not psxbuttons(5);
	
	cpu_reset_h<=reset_h or psxbuttons(0);
	reset_h<=I_RESET;
	reset_l<=not I_RESET;
	xval_inv<=xval xor "1000000000";
	yval_inv<=yval xor "1000000000";
	
--	yval_pwm<=yval_inv when psxbuttons(8)='1' else yval_filt;
--	xval_pwm<=xval_inv when psxbuttons(8)='1' else xval_filt;
	yval_pwm<=xval_inv; --board swaps x andf y
	xval_pwm<=yval_inv;
--	yval_pwm<=xval_filt; --board swaps x andf y
--	xval_pwm<=yval_filt;

	O_VIDEO_Z<=zval(7 downto 4);

	O_AUDIO_L<=soundout;
	O_AUDIO_R<=soundout;

	O_RDIS<=not rgb(0);
	O_GDIS<=not rgb(1);
	O_BDIS<=not rgb(2);
--	O_RDIS<=rgb(0);
--	O_GDIS<=rgb(1);
--	O_BDIS<=rgb(2);

end Behavioral;
