/*
 * main.c: Atari Vector game simulator
 *
 * Copyright 1991-1993, 1996, 2003 Hedley Rainnie, Doug Neubauer, and Eric Smith
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: main.c 2 2003-08-20 01:51:05Z eric $
 */

#ifndef __GNUC__
#define inline
#endif

#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>

#ifdef USE_HISTORY
#include <readline/history.h>
#endif

#include "memory.h"
#include "game.h"
#include "display.h"
#include "sim6502.h"
#include "debugger.h"


int getpid(void);


void
foo()
{
    stepflag = 1;
    traceflag = 0;
    signal(SIGINT, foo);
}



void show_usage_message (char *av0)
{
#ifdef SINGLE_GAME
  fprintf (stderr, "Usage: %s [<options>]\n", av0);
#else
  fprintf (stderr, "Usage: %s <game> [<options>]\n", av0);
#endif
  fprintf (stderr, 
	   "   -st     self test mode\n"
#ifdef VG_DEBUG
	   "   -vgo    trace vgo's\n"
#endif
	   "   -load <file>  specify a memory dump file to load\n"
	   "   -df <n> display only every nth frame\n"
	   "   -large  use a large window\n"
	   "   -nopm   don't use a pixmap (faster?)\n"
	   "   -wide   use wide lines\n"
	   "   -brk    break into debugger before starting\n"
	   );
#ifndef SINGLE_GAME
  fprintf (stderr,
	   "\nThe following games are supported:\n"
	   "    keyword       title\n"
	   "    ----------    ------------------------------\n");
  show_games ();
#endif
}


int main(int argc, char *argv[])
{
  int ac;
  char **av;

  int show_usage = 1;
  char *reload_file = NULL;

  int smallwindow = 1;
  int use_pixmap = 1;
  int line_width = 0;

#ifdef THINK_C
  c = ccommand(&argv);
#endif

#ifdef USE_HISTORY
  using_history();
#endif

  mem = (elem *) calloc(65536L,sizeof(elem));
  if(!mem)
    {
      printf("Cannot get memory for simulation\n");
      exit(1);
    }

  ac = argc;
  av = argv;

  while (argc > 1)
    {
      argc--; argv++;
      if (*argv[0] == '-')
	{
	  if (strcasecmp (argv [0], "-st") == 0)
	    self_test = 1;
	  else if (strcasecmp (argv [0], "-load") == 0)
	    {
	      argc--; argv++;
	      if (reload_file == NULL)
		reload_file = argv [0];
	      else
		{
		  fprintf (stderr, "Only one load file may be specified\n");
		  goto quit;
		}
	    }
#ifdef VG_DEBUG
	  else if (strcasecmp (argv [0], "-vgo") == 0)
	    trace_vgo = 1;
#endif
	  else if (strcasecmp (argv [0], "-df") == 0)
	    {
	      argc--; argv++;
	      drop_frames = atoi (argv [0]);
	    }
	  else if (strcasecmp (argv [0], "-large") == 0)
	    smallwindow = 0;
	  else if (strcasecmp (argv [0], "-nopm") == 0)
	    use_pixmap = 0;
	  else if (strcasecmp (argv [0], "-wide") == 0)
	    line_width = 2;
	  else if (strcasecmp (argv [0], "-game") == 0)
	    ; /* obsolete */
	  else if (strcasecmp (argv [0], "-brk") == 0)
	    stepflag = 1;
	  else if (strcasecmp (argv [0], "-help") == 0)
	    goto quit;
	  else if (strcasecmp (argv [0], "-?") == 0)
	    goto quit;
#ifdef COUNT_INTERRUPTS
	  else if (strncasecmp (argv [0], "-intquit", 8) == 0)
	    int_quit = atoi (argv [0] + 8);
#endif
	  else
	    {
	      fprintf (stderr, "Unrecognized option \"%s\"\n", argv[0]);
	      goto quit;
	    }
	}
      else
	{
#ifdef SINGLE_GAME
	  fprintf (stderr, "Unrecognized option \"%s\"\n", argv[0]);
	  goto quit;
#else
	  if (game == 0)
	    game = pick_game (argv [0]);
	  else
	    {
	      fprintf (stderr, "Only one game may be specified\n");
	      goto quit;
	    }
#endif
	}
    }

#ifdef SINGLE_GAME
  game = pick_game (SINGLE_GAME);
#else
  if (game == 0)
    {
      fprintf (stderr, "A game must be specified\n");
      goto quit;
    }
#endif

  show_usage = 0;

  signal(SIGINT, foo);

  srand(getpid());

  if (reload_file != NULL)
    reload (reload_file);
  else
    {
      setup_game ();
      save_PC = (memrd(0xfffd,0,0) << 8) | memrd(0xfffc,0,0);
      save_A = 0;
      save_X = 0;
      save_Y = 0;
      save_flags = 0;
      save_totcycles = 0;
      irq_cycle = 8192;
    }

  init_graphics (argc, argv, smallwindow, use_pixmap, line_width, game_name (game));

  sim_6502 ();

 quit:
  if (show_usage)
    {
      show_usage_message (av [0]);
      return 1;
    }
  return 0;
}
